/*
 * Copyright 2012 Jan Ouwens
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package nl.jqno.equalsverifier;

import static nl.jqno.equalsverifier.testhelpers.Util.assertFailure;
import static nl.jqno.equalsverifier.testhelpers.Util.nullSafeEquals;
import static nl.jqno.equalsverifier.testhelpers.Util.nullSafeHashCode;
import static org.junit.Assert.assertEquals;
import static org.junit.Assert.assertTrue;
import nl.jqno.equalsverifier.MutableStateTest.MutableIntContainer;
import nl.jqno.equalsverifier.testhelpers.MockStaticFieldValueStash;
import nl.jqno.equalsverifier.util.FieldAccessor;
import nl.jqno.equalsverifier.util.ObjectAccessor;

import org.junit.Test;

public class OriginalStateTest {
	private static final Object INSTANCE_1 = new Object();
	private static final Object INSTANCE_2 = new Object();
	private static final Object STATIC = new Object();
	private static final Object STATIC_FINAL = new Object();
	
	@Test
	public void staticValueReturnsToOriginalState() {
		EqualsVerifier.forClass(CorrectEquals.class).verify();
		assertEquals(STATIC_FINAL, CorrectEquals.staticFinalValue);
		assertEquals(STATIC, CorrectEquals.staticValue);
	}
	
	@Test
	public void instanceValueReturnsToOriginalState() {
		CorrectEquals one = new CorrectEquals(INSTANCE_1);
		CorrectEquals two = new CorrectEquals(INSTANCE_2);
		EqualsVerifier.forExamples(one, two).verify();
		
		assertEquals(INSTANCE_1, one.instanceValue);
		assertEquals(INSTANCE_2, two.instanceValue);
	}
	
	@Test
	public void staticValueReturnsToOriginalStateRecursively() {
		EqualsVerifier.forClass(CorrectEqualsContainer.class).verify();
		assertEquals(STATIC, CorrectEquals.staticValue);
	}
	
	@Test
	public void staticValueReturnsToOriginalStateDeeplyRecursively() {
		EqualsVerifier.forClass(CorrectEqualsContainerContainer.class).verify();
		assertEquals(STATIC, CorrectEquals.staticValue);
	}

	@Test
	public void staticValueInSuperReturnsToOriginalState() {
		EqualsVerifier.forClass(SubContainer.class).verify();
		assertEquals(STATIC, CorrectEquals.staticValue);
		assertEquals(STATIC, SuperContainer.staticValue);
		assertEquals(STATIC_FINAL, SuperContainer.staticFinalValue);
	}
	
	@Test
	public void valuesReturnToOriginalStateAfterException() throws NoSuchFieldException {
		EqualsVerifier<MutableIntContainer> ev = EqualsVerifier.forClass(MutableIntContainer.class);
		MockStaticFieldValueStash stash = new MockStaticFieldValueStash();

		// Mock EqualsVerifier's StaticFieldValueStash
		ObjectAccessor<EqualsVerifier<MutableIntContainer>> objectAccessor = ObjectAccessor.of(ev);
		FieldAccessor stashAccessor = objectAccessor.fieldAccessorFor(EqualsVerifier.class.getDeclaredField("stash"));
		stashAccessor.set(stash);

		// Make sure the exception actually occurs, on a check that actually mutates the fields.
		assertFailure(ev, "Mutability");

		// Assert
		assertTrue(stash.restoreCalled);
	}
	
	static final class CorrectEquals {
		private static final Object staticFinalValue = STATIC_FINAL;
		private static Object staticValue = STATIC;
		private final Object instanceValue;
		
		public CorrectEquals(Object instanceValue) {
			this.instanceValue = instanceValue;
		}
		
		@Override
		public boolean equals(Object obj) {
			if (!(obj instanceof CorrectEquals)) {
				return false;
			}
			CorrectEquals other = (CorrectEquals)obj;
			return nullSafeEquals(instanceValue, other.instanceValue);
		}
		
		@Override
		public int hashCode() {
			return nullSafeHashCode(instanceValue);
		}
	}
	
	static final class CorrectEqualsContainer {
		private final CorrectEquals foo;
		
		public CorrectEqualsContainer(CorrectEquals foo) {
			this.foo = foo;
		}
		
		@Override
		public boolean equals(Object obj) {
			if (!(obj instanceof CorrectEqualsContainer)) {
				return false;
			}
			CorrectEqualsContainer other = (CorrectEqualsContainer)obj;
			return nullSafeEquals(foo, other.foo);
		}
		
		@Override
		public int hashCode() {
			return nullSafeHashCode(foo);
		}
	}
	
	static final class CorrectEqualsContainerContainer {
		private final CorrectEqualsContainer foo;
		
		public CorrectEqualsContainerContainer(CorrectEqualsContainer foo) {
			this.foo = foo;
		}
		
		@Override
		public boolean equals(Object obj) {
			if (!(obj instanceof CorrectEqualsContainerContainer)) {
				return false;
			}
			CorrectEqualsContainerContainer other = (CorrectEqualsContainerContainer)obj;
			return nullSafeEquals(foo, other.foo);
		}
		
		@Override
		public int hashCode() {
			return nullSafeHashCode(foo);
		}
	}
	
	static abstract class SuperContainer {
		private static final Object staticFinalValue = STATIC_FINAL;
		private static Object staticValue = STATIC;
		
		private final CorrectEquals foo;
		
		public SuperContainer(CorrectEquals foo) {
			this.foo = foo;
		}
		
		@Override
		public boolean equals(Object obj) {
			if (!(obj instanceof SuperContainer)) {
				return false;
			}
			SuperContainer other = (SuperContainer)obj;
			return nullSafeEquals(foo, other.foo);
		}
		
		@Override
		public int hashCode() {
			return nullSafeHashCode(foo);
		}
	}

	static final class SubContainer extends SuperContainer {
		public SubContainer(CorrectEquals foo) {
			super(foo);
		}
	}
}
